import pytest
from gusty import create_dag
from datetime import timedelta

###############
## FIXTURES ##
###############


@pytest.fixture(scope="session")
def external_deps_dir():
    return "tests/dags/external_deps"


@pytest.fixture(scope="session")
def dag(external_deps_dir):
    dag = create_dag(
        external_deps_dir,
        default_args={"email": "default@gusty.com", "retries": 5},
        task_group_defaults={"prefix_group_id": True},
        wait_for_defaults={"poke_interval": 12},
    )
    return dag


###########
## TESTS ##
###########


def test_dag_level_ext_dep(dag):
    wait_for_task = dag.task_dict["wait_for_DAG_top_level_external"]
    assert wait_for_task.__dict__["execution_delta"] == timedelta(minutes=88)


def test_task_group_level_ext_dep(dag):
    wait_for_task = dag.task_dict["wait_for_some_task"]
    assert wait_for_task.__dict__["execution_delta"] == timedelta(minutes=89)


def test_task_level_ext_dep(dag):
    wait_for_task = dag.task_dict["wait_for_another_task_1"]
    assert wait_for_task.__dict__["execution_delta"] == timedelta(minutes=95)
    assert wait_for_task.__dict__["retries"] == 95



import pytest
import os
from airflow import DAG
from datetime import datetime, timedelta
from airflow.utils.dates import days_ago
from gusty import create_dag


# This module is intended to test the default behavior of gusty
# Given a directory of tasks without any metadata located in those directories
# A sort of baseline


##############
## FIXTURES ##
##############


@pytest.fixture(scope="session")
def no_metadata_dir():
    return "tests/dags/ignore_subfolders"


@pytest.fixture(scope="session")
def dag(no_metadata_dir):
    dag = create_dag(
        no_metadata_dir,
        description="A dag created without metadata",
        schedule_interval="0 0 * * *",
        default_args={
            "owner": "gusty",
            "depends_on_past": False,
            "start_date": days_ago(1),
            "email": "gusty@gusty.com",
            "email_on_failure": False,
            "email_on_retry": False,
            "retries": 3,
            "retry_delay": timedelta(minutes=5),
        },
        ignore_subfolders=True,
    )
    return dag


###########
## TESTS ##
###########


def test_ignore_subfolders(dag):
    assert "ignore_this_task" not in dag.task_dict.keys()


def test_ignore_skiplevel(dag):
    assert "ignore_this_too" not in dag.task_dict.keys()



import pytest
from gusty import create_dag
from airflow.operators.dummy import DummyOperator

###############
## FIXTURES ##
###############


@pytest.fixture(scope="session")
def pydag_dir():
    return "tests/dags/pydag"


@pytest.fixture(scope="session")
def dag(pydag_dir):
    dag = create_dag(
        pydag_dir,
        default_args={"email": "default@gusty.com", "retries": 5},
        task_group_defaults={"prefix_group_id": True},
        wait_for_defaults={"poke_interval": 12},
    )
    return dag


@pytest.fixture(scope="session")
def py_task(dag):
    py_task = dag.task_dict["py_task"]
    return py_task


###########
## TESTS ##
###########


def test_py_task_exists(dag):
    assert "py_task" in dag.task_dict.keys()


def test_py_task_callable(py_task):
    callable = py_task.__dict__["python_callable"]
    res = callable()
    assert res == "hello python task"


def test_py_task_overrides(py_task):
    assert py_task.__dict__["email"] == "new_email@gusty.com"


def test_py_task_external_dependencies(py_task):
    assert "wait_for_DAG_a_whole_dag" in py_task.__dict__["_upstream_task_ids"]


def test_py_task_dependencies(py_task):
    assert "direct_dep" in py_task.__dict__["_upstream_task_ids"]
    assert "simple_leaf" in py_task.__dict__["_downstream_task_ids"]


def test_py_dummy(dag):
    """
    This really tests to ensure that in theory the contents of the .py could
    be run by other operators as needed.
    """
    task = dag.task_dict["py_but_dummy"]
    assert isinstance(task, DummyOperator)


def test_automatic_py(dag, py_task):
    task = dag.task_dict["simple_leaf"]
    assert type(py_task) == type(task)



import pytest
import os
from airflow import DAG
from airflow.models import BaseOperator
from datetime import datetime, timedelta
from airflow.utils.dates import days_ago
from gusty.building import create_schematic
from gusty.parsing import default_parsers
from gusty import create_dag


# This module is intended to test the default behavior of gusty
# Given a directory of tasks without any metadata located in those directories
# A sort of baseline


##############
## FIXTURES ##
##############


@pytest.fixture(scope="session")
def no_metadata_dir():
    return "tests/dags/no_metadata"


@pytest.fixture(scope="session")
def dag_files(no_metadata_dir):
    schematic = create_schematic(no_metadata_dir)
    dag_file_list = [level["spec_paths"] for level_id, level in schematic.items()]
    dag_files = [item for sublist in dag_file_list for item in sublist]
    return dag_files


@pytest.fixture(scope="session")
def dag(no_metadata_dir):
    dag = create_dag(
        no_metadata_dir,
        description="A dag created without metadata",
        schedule_interval="0 0 * * *",
        default_args={
            "owner": "gusty",
            "depends_on_past": False,
            "start_date": days_ago(1),
            "email": "gusty@gusty.com",
            "email_on_failure": False,
            "email_on_retry": False,
            "retries": 3,
            "retry_delay": timedelta(minutes=5),
        },
        external_dependencies=[{"external_dag": "a_root_level_external"}],
    )
    return dag


###########
## TESTS ##
###########


def test_dag_is_dag(dag):
    assert isinstance(dag, DAG)


def test_dag_named_after_dir(dag, no_metadata_dir):
    assert dag.dag_id == os.path.basename(no_metadata_dir)


def test_latest_only_root(dag):
    assert len(dag.roots) == 1
    assert dag.roots[0].task_id == "latest_only"


def test_task_named_after_file(dag):
    top_level_task = dag.task_dict["top_level_task"]
    assert top_level_task is not None


def test_default_no_task_group_prefixes(dag):
    assert "task_overrides_retries" in dag.task_dict.keys()


def test_ipynb_task_exists(dag):
    assert "dependent_jupyter_task" in dag.task_dict.keys()


def test_task_overrides_defaults(dag):
    override_task = dag.task_dict["task_overrides_retries"]
    assert override_task.__dict__["retries"] > dag.default_args["retries"]


def test_top_level_depends_on_second_level(dag):
    top_level_task = dag.task_dict["top_level_task"]
    top_level_task_dependencies = top_level_task.__dict__["_upstream_task_ids"]
    assert "dependable_task" in top_level_task_dependencies


def test_external_dependencies(dag):
    dependable_task = dag.task_dict["dependable_task"]
    dependable_task_dependencies = dependable_task.__dict__["_upstream_task_ids"]
    assert "wait_for_DAG_a_whole_dag" in dependable_task_dependencies
    assert "wait_for_external_task" in dependable_task_dependencies
    assert "wait_for_external_task_2" in dependable_task_dependencies


def test_external_dependencies_dict_tasks(dag):
    ext_dep_dict_spec_task = dag.task_dict["task_overrides_retries"]
    ext_dep_dict_spec_task_dependencies = ext_dep_dict_spec_task.__dict__[
        "_upstream_task_ids"
    ]
    assert "wait_for_DAG_a_whole_dag" in ext_dep_dict_spec_task_dependencies
    assert "wait_for_another_task_1" in ext_dep_dict_spec_task_dependencies
    assert "wait_for_another_task_2" in ext_dep_dict_spec_task_dependencies


def test_tasks_created(dag, dag_files):
    def replace_extension(file):
        final = None
        for extension in default_parsers.keys():
            if file.endswith(extension):
                final = file.replace(extension, "")
        return final

    expected_tasks = [*map(os.path.basename, dag_files)]
    expected_tasks = [*map(replace_extension, expected_tasks)]
    tasks_present = [task in dag.task_dict.keys() for task in expected_tasks]
    tasks_are_operators = [
        isinstance(dag.task_dict[task], BaseOperator) for task in expected_tasks
    ]
    assert all(expected_tasks)
    assert all(tasks_are_operators)


def test_args_passed(dag):
    top_level_task = dag.task_dict["top_level_task"]
    assert top_level_task.__dict__["bash_command"] == "echo hello"


def test_sensor_support(dag):
    assert dag.task_dict["sensor_task"].__dict__["delta"] == timedelta(seconds=2)


def test_root_external_dependencies_accepted(dag):
    assert (
        "wait_for_a_root_level_external"
        in dag.task_dict["latest_only"].__dict__["_downstream_task_ids"]
    )


def test_root_external_dependencies_latest_only_order(dag):
    # Bad test name, but bascially nothing should be upstream of latest only,
    # And only the DAG-level external dependency is after latest_only
    latest_only = dag.task_dict["latest_only"].__dict__
    latest_only_downstream_tasks = latest_only["_downstream_task_ids"]
    latest_only_upstream_tasks = latest_only["_upstream_task_ids"]
    assert (
        len(latest_only_upstream_tasks) == 0 and len(latest_only_downstream_tasks) == 1
    )



import yaml
from datetime import datetime, timedelta
from airflow.utils.dates import days_ago


def scalar_to_value(scalar):
    """
    Converts a YAML ScalarNode to its underlying Python value
    """
    type = scalar.tag.split(":")[-1]
    val = scalar.value
    return eval("{type}('{val}')".format(type=type, val=val))


def node_converter(x):
    """
    Converts YAML nodes of varying types into Python values,
    lists, and dictionaries
    """
    if isinstance(x, yaml.ScalarNode):
        # "I am an atomic value"
        return yaml.load(x.value, yaml.SafeLoader)
    if isinstance(x, yaml.SequenceNode):
        # "I am a list"
        return [scalar_to_value(v) for v in x.value]
    if isinstance(x, yaml.MappingNode):
        # "I am a dict"
        return {scalar_to_value(v[0]): scalar_to_value(v[1]) for v in x.value}


def wrap_yaml(func):
    """Turn a function into one that can be run on a YAML input"""

    def ret(loader, x):
        value = node_converter(x)

        if value is not None:

            if isinstance(value, list):
                return func(*value)

            if isinstance(value, dict):
                return func(**value)

            return func(value)

        else:
            return func()

    return ret


def generate_loader(custom_constructors={}):
    """Generates a SafeLoader with both standard Airflow and custom constructors"""
    loader = yaml.SafeLoader
    dag_yaml_tags = {
        "!days_ago": days_ago,
        "!timedelta": timedelta,
        "!datetime": datetime,
    }

    if isinstance(custom_constructors, list) and len(custom_constructors) > 0:
        custom_constructors = {
            ("!" + func.__name__): func for func in custom_constructors
        }

    if len(custom_constructors) > 0:
        dag_yaml_tags.update(custom_constructors)
    for tag, func in dag_yaml_tags.items():
        loader.add_constructor(tag, wrap_yaml(func))
    return loader



import os
import inspect
from gusty.parsing.loaders import generate_loader
from gusty.parsing.parsers import parse_generic, parse_py, parse_ipynb, parse_sql

default_parsers = {
    ".yml": parse_generic,
    ".yaml": parse_generic,
    ".Rmd": parse_generic,
    ".py": parse_py,
    ".ipynb": parse_ipynb,
    ".sql": parse_sql,
}


def parse(file_path, parse_dict=default_parsers, loader=None):
    """
    Reading in yaml specs / frontmatter.
    """

    if loader is None:
        loader = generate_loader()

    path, extension = os.path.splitext(file_path)

    parser = parse_dict[extension]

    if "loader" in inspect.signature(parser).parameters.keys():
        yaml_file = parser(file_path, loader=loader)
    else:
        yaml_file = parser(file_path)

    assert "operator" in yaml_file, "No operator specified in yaml spec " + file_path

    # gusty always supplies a task_id and a file_path in a spec
    yaml_file["task_id"] = (
        os.path.splitext(os.path.basename(file_path))[0].lower().strip()
    )
    assert (
        yaml_file["task_id"] != "all"
    ), "Task name 'all' is not allowed. Please change your task name."

    yaml_file["file_path"] = file_path

    # Check dependencies
    if "dependencies" in yaml_file.keys():
        assert isinstance(
            yaml_file["dependencies"], list
        ), "dependencies needs to be a list of strings in {file_path}".format(
            file_path=file_path
        )
        assert all(
            [isinstance(dep, str) for dep in yaml_file["dependencies"]]
        ), "dependencies needs to be a list of strings in {file_path}".format(
            file_path=file_path
        )

    return yaml_file



import yaml, ast, importlib.util, nbformat, jupytext, re
from gusty.parsing.loaders import generate_loader
from gusty.importing import airflow_version

if airflow_version > 1:
    from airflow.operators.python import PythonOperator
else:
    from airflow.operators.python_operator import PythonOperator


def frontmatter_load(file_path, loader=None):
    """
    Loads YAML frontmatter. Expects a YAML block at the top of the file
    that starts and ends with "---". In use in favor of frontmatter.load
    so that custom dag_constructors (via PyYaml) can be used uniformly across
    all file types.
    """
    if loader is None:
        loader = generate_loader()
    FM_BOUNDARY = re.compile(r"^-{3,}\s*$", re.MULTILINE)
    with open(file_path, "r") as file:
        text = "".join(file.readlines())
        if text.startswith("---"):
            _, metadata, content = FM_BOUNDARY.split(text, 2)
            metadata = yaml.load(metadata, Loader=loader)
            content = content.strip("\n")
        else:
            metadata = None
            content = yaml.load(text, Loader=loader)
    return {"metadata": metadata, "content": content}


def parse_generic(file_path, loader=None):
    if loader is None:
        loader = generate_loader()
    # Read either the frontmatter or the parsed yaml file (using "or" to coalesce them)
    file_contents = frontmatter_load(file_path)
    job_spec = file_contents["metadata"] or file_contents["content"]

    return job_spec


def parse_py(file_path, loader=None):
    if loader is None:
        loader = generate_loader()

    job_spec = {}
    if airflow_version > 1:
        job_spec.update({"operator": "airflow.operators.python.PythonOperator"})
    else:
        job_spec.update(
            {"operator": "airflow.operators.python_operator.PythonOperator"}
        )

    file_contents = jupytext.read(file_path)["cells"][0]

    # if spec contains metadata header...
    if file_contents["cell_type"] == "raw":
        assert (
            file_contents["source"] is not None
        ), "You need a comment block starting and ending with '# ---' at the top of {file_path}".format(
            file_path=file_path
        )
        assert (
            "---" in file_contents["source"],
        ), "You need a comment block starting and ending with '# ---' at the top of {file_path}".format(
            file_path=file_path
        )
        source = file_contents["source"].replace("---", "")
        settings = yaml.load(source, Loader=loader)
        job_spec.update(**settings)

        # search for a python callable if one is specified
        if "python_callable" in job_spec.keys():
            with open(file_path) as f:
                tree = ast.parse(f.read())

                class Visitor(ast.NodeVisitor):
                    def __init__(self):
                        self.has_callable = None

                    def visit_FunctionDef(self, node):
                        ast.NodeVisitor.generic_visit(self, node)
                        if node.name == job_spec["python_callable"]:
                            self.has_callable = True

                v = Visitor()
                v.visit(tree)
                if v.has_callable:
                    mod_file = importlib.util.spec_from_file_location(
                        "".join(i for i in file_path if i.isalnum()), file_path
                    )
                    mod = importlib.util.module_from_spec(mod_file)
                    mod_file.loader.exec_module(mod)
                    job_spec.update(
                        {"python_callable": getattr(mod, job_spec["python_callable"])}
                    )
                else:
                    assert (
                        False
                    ), "{file_path} specifies python_callable {callable} but {callable} not found in {file_path}".format(
                        file_path=file_path, callable=job_spec["python_callable"]
                    )
        # Default to sourcing this file for a PythonOperator
        else:
            job_spec.update({"python_callable": lambda: exec(open(file_path).read())})
    # If no metadata then we also default to sourcing this file for a PythonOperator
    else:
        job_spec.update({"python_callable": lambda: exec(open(file_path).read())})

    return job_spec


def parse_ipynb(file_path, loader=None):
    if loader is None:
        loader = generate_loader()
    # Find first yaml cell in jupyter notebook and parse yaml
    file_contents = nbformat.read(file_path, as_version=4)["cells"]
    yaml_cell = [
        cell
        for cell in file_contents
        if cell["cell_type"] == "markdown"
        and cell["source"].startswith(("```yaml", "```yml"))
    ]
    assert len(yaml_cell) > 0, "Please add a yaml block to %s" % file_path
    yaml_cell = yaml_cell[0]["source"]
    job_spec = yaml.load(
        yaml_cell.replace("```yaml", "").replace("```yml", "").replace("```", ""),
        Loader=loader,
    )

    return job_spec


def parse_sql(file_path, loader=None):
    if loader is None:
        loader = generate_loader()
    file_contents = frontmatter_load(file_path, loader=loader)
    job_spec = file_contents["metadata"]
    job_spec["sql"] = file_contents["content"]
    return job_spec



import airflow
from gusty.building import GustyBuilder


def create_dag(
    dag_dir,
    task_group_defaults={},
    wait_for_defaults={},
    latest_only=True,
    parse_hooks={},
    dag_constructors={},
    **kwargs
):
    """
    Create an Airflow DAG by passing a directory of .yml task files.

    Each .yml task file should represent one task in the DAG. .ipynb and .Rmd files with yaml frontmatter are also
    accepted by create_dag. Any DAG kwargs (e.g. schedule_interval, description) can be passed directly into create_dag.
    Additionally, if your directory contains multiple subdirectories of .yml task files, and you are on Airflow V2, create_dag
    will create task groups for each subdirectory. By default, prefix_group_id is set to False.

    Parameters:
        dag_dir (str): A path to a directory of .yml task files. .ipynb and .Rmd files with yaml frontmatter are also accepted
        task_group_defaults (dict): Any parameter that could be passed to Airflow's TaskGroup class
        wait_for_defaults (dict): Any parameter that could be passed to Airflow's ExternalTaskSensor (which is what gusty users to create external_dependency tasks)
        latest_only (bool): When True, creates a latest_only operator at the root of the DAG, which will ensure tasks only run for the most recent run date
        parse_hooks (dict): Keys are file extensions (beginning with periods), values are functions that parse a given file path for a the associated file extension.
        kwargs: Any additional keyword argument that can be passed to Airflow's DAG class, e.g. default_args, schedule_interval, description

    Returns:
        dag (airflow.DAG): A fully initialized Airflow DAG, complete with tasks, task groups, dependencies, external dependencies, and latest_only operator (if wanted).
    """

    setup = GustyBuilder(
        dag_dir,
        task_group_defaults=task_group_defaults,
        wait_for_defaults=wait_for_defaults,
        latest_only=latest_only,
        parse_hooks=parse_hooks,
        dag_constructors=dag_constructors,
        **kwargs
    )
    [setup.parse_metadata(level) for level in setup.levels]
    [setup.check_metadata(level) for level in setup.levels]
    [setup.create_structure(level) for level in setup.levels]
    [setup.read_specs(level) for level in setup.levels]
    [setup.create_tasks(level) for level in setup.levels]
    [setup.create_level_dependencies(level) for level in setup.levels]
    [setup.create_task_dependencies(level) for level in setup.levels]
    [setup.create_task_external_dependencies(level) for level in setup.levels]
    [setup.create_level_external_dependencies(level) for level in setup.levels]
    [setup.create_leaf_tasks(level) for level in setup.levels]
    [setup.create_root_dependencies(level) for level in setup.levels]
    return setup.return_dag()



from gusty.importing import airflow_version

if airflow_version > 1:
    from airflow.sensors.external_task import ExternalTaskSensor
else:
    from airflow.sensors.external_task_sensor import ExternalTaskSensor


def create_external_sensor(
    dag,
    external_dag_id,
    external_task_id,
    wait_for_defaults,
    wait_for_overrides,
    wait_for_task_name,
    sensor=ExternalTaskSensor,
):
    ext_task_defaults = wait_for_defaults.copy()
    ext_task_overrides = wait_for_overrides.get(external_dag_id, {})
    # Only one of execution_delta or
    # execution_date_fn can be passed to
    # the ExternalTaskSensor
    if "execution_delta" in ext_task_overrides:
        if "execution_date_fn" in ext_task_defaults:
            ext_task_defaults.pop("execution_date_fn")
    ext_task_defaults.update(ext_task_overrides)
    wait_for_task = sensor(
        dag=dag,
        task_id=wait_for_task_name,
        external_dag_id=external_dag_id,
        external_task_id=(external_task_id if external_task_id != "all" else None),
        **ext_task_defaults
    )

    return wait_for_task


def make_external_task_name(external_dag_id, external_task_id):
    return (
        "wait_for_DAG_{x}".format(x=external_dag_id)
        if external_task_id == "all"
        else "wait_for_{x}".format(x=external_task_id)
    )



from airflow.models.baseoperator import BaseOperator
from gusty.building import _get_operator_parameters


def test_get_operator_parameters():
    class ACustomOperator(BaseOperator):
        def __init__(self, a, **kwargs):
            self.a = 1

            super().__init__(**kwargs)

        def execute(self, context):
            print(self.a)

    params = _get_operator_parameters(ACustomOperator)

    assert "a" in params
    assert list(params) == ["self", "a", "kwargs"]


def test_get_operator_parameters_attribute():
    f = lambda a, **kwargs: BaseOperator(**kwargs)
    f._gusty_parameters = ("a",)

    params = _get_operator_parameters(f)

    assert "a" in params
    assert list(params) == ["a"]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gusty",
    version="0.8.2",
    author="Chris Cardillo, Michael Chow, David Robinson",
    author_email="cfcardillo23@gmail.com",
    description="Making DAG construction easier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/chriscardillo/gusty",
    packages=setuptools.find_packages(),
    install_requires=["apache-airflow", "inflection", "jupytext", "nbformat", "PyYaml"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import pytest
from gusty import create_dag
from datetime import datetime, timedelta
from airflow.utils.dates import days_ago

###############
## FIXTURES ##
###############


@pytest.fixture(scope="session")
def task_group_dependencies_dir():
    return "tests/dags/task_group_dependencies"


@pytest.fixture(scope="session")
def dag(task_group_dependencies_dir):
    dag = create_dag(
        task_group_dependencies_dir,
        description="A dag created without metadata",
        schedule_interval="0 0 * * *",
        default_args={
            "owner": "gusty",
            "depends_on_past": False,
            "start_date": days_ago(1),
            "email": "gusty@gusty.com",
            "email_on_failure": False,
            "email_on_retry": False,
            "retries": 3,
            "retry_delay": timedelta(minutes=5),
        },
        leaf_tasks=["simple_leaf"],
    )
    return dag


###########
## TESTS ##
###########


def test_task_group_dependencies(dag):
    """
    This test is to ensure that, when root tasks are created (e.g. latest_only),
    task groups which depend on other tasks groups are not considered valid dependency
    objects for root tasks.
    """
    second_task_group = dag.__dict__["_task_group"].__dict__["children"][
        "second_task_group"
    ]
    assert "latest_only" not in second_task_group.upstream_task_ids
